from construct import *
import struct

def round_up(n, m):
    return ((n - 1) / m + 1) * m

def round_down(n, m):
    return n / m * m

try:
    # raise ImportError
    from _algos import parsebitmap 
except ImportError:
    import warnings
    warnings.warn('could not load _algos, falling back to Python implementation')
    def parsebitmap(stream, context):
        from array import array
        w = context.w
        h = context.h
        bmpw = round_up(w, 4)
        srcw = round_down(w, 4)
        # TODO find out how sogna really handles widths that aren't multiples of four... fill, offset, what?
        output = array('B', [0]*bmpw*h)
        x, y, repeat = 0, 0, 0
        last = None
        
        while x < srcw:
            cur = array('B', stream.read(4))
            
            if not repeat:
                repeat = 1
            elif cur == last:
                repeat += 1
            
            j = bmpw*(h-y-1)+x
            output[j:j+4] = cur
            y += 1
            
            if y == h: 
                y = 0
                x += 4
                if repeat == 1:
                    repeat = 0
            
            if repeat:
                last = cur
            
            if repeat == 2:
                m = ord(stream.read(1))
                if not m:
                    m = 256 + ord(stream.read(1))
                assert m
                
                until = m + y - 1
                while y < until:
                    j = bmpw*(h-y-1)+x
                    output[j:j+4] = cur
                    y += 1
                if y == h:
                    y = 0
                    x += 4
                repeat = 0
        
        return output.tostring()

class CompressedBitmap(Construct):
    __slots__ = []
    def __init__(self, name):
        Construct.__init__(self, name)
    def _parse(self, stream, context):
        # print 'parsing bitmap'
        data = parsebitmap(stream, context)
        # print 'bitmap parsed'
        return data
        
    def _build(self, obj, stream, context):
        raise NotImplementedError
    def _sizeof(self, context):
        raise SizeofError("can't calculate size")

class UncompressedBitmap(Adapter):
    def _encode(self, obj, context):
        raise NotImplementedError
    def _decode(self, obj, context):
        w = context.w
        h = context.h
        if w % 4:
            raise NotImplementedError
        rows = []
        for y in xrange(h-1,-1,-1):
            rows.append(obj[w*y:w*y+w])
        return ''.join(rows)

Animation = Struct('animation',
    StrictRepeater(256, Struct('palette',
        ULInt8('blue'), ULInt8('green'), ULInt8('red')
    )),
    ULInt32('frame_count'),
    Value('uncompressed', lambda ctx: bool(ctx.frame_count >> 31)),
    Value('frame_count', lambda ctx: ctx.frame_count & 0x7FFFFFFF),
    Anchor('offset_start'),
    MetaRepeater(lambda ctx: ctx['frame_count'],
        Struct('offsets',
            ULInt32('offset'),
            OnDemandPointer(lambda ctx: ctx._.offset_start + ctx.offset, Struct('frame',
                ULInt16('x'),
                ULInt16('y'),
                ULInt16('w'),
                ULInt16('h'),
                IfThenElse('data', lambda ctx: ctx._._.uncompressed,
                    UncompressedBitmap(Field(None, lambda ctx: ctx.w * ctx.h)),
                    CompressedBitmap(None)
                )
            ))
        )
    )
)

def loadframes(data):
    anm = Animation.parse(data)
    return anm.palette, [offset.frame.value for offset in anm.offsets]

def writebmp(frame, palette, out):
    out.write('BM')
    out.write(struct.pack('III', len(frame.data)+1078, 0, 1078))
    out.write(struct.pack('IIIHHI', 40, frame.w, frame.h, 1, 8, 0))
    out.write(struct.pack('IIIII', 0, 1, 1, 0, 0))
    for i, color in enumerate(palette):
        out.write(struct.pack('BBBB', color.blue, color.green, color.red, 255))
    out.write(frame.data)
    return out

if __name__ == '__main__':
    import sys, os
    from sogna.dat import DATArchive
    from traceback import print_exc
    from fnmatch import filter as fnfilter

    def extract(dat, name, dest):
        data = dat.read(name)
        anm = Animation.parse(data)
        dirpath = os.path.dirname(os.path.join(dest, name))
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        for index, offset in enumerate(anm.offsets):
            fn = '%s[%d].BMP'%(name, index+1)
            path = os.path.join(dest, fn)
            if os.path.exists(path):
                continue
            try:
                frame = offset.frame.value
            except Exception:
                print_exc()
                with open(os.path.join(dest, name), 'wb') as o:
                    o.write(data)
                break # continue
            with open(path, 'wb') as o:
                writebmp(frame, anm.palette, o)

    with DATArchive(sys.argv[1]) as dat:
        if len(sys.argv) < 3:
            for fn in fnfilter(dat.filenames(), '*.ANM'):
                print fn
        elif sys.argv[2] == '*':
            for fn in fnfilter(dat.filenames(), '*.ANM'):
                print fn
                extract(dat, fn, sys.argv[3])
        elif '*' in sys.argv[2] or '?' in sys.argv[2]:
            names = fnfilter(dat.filenames(), '*.ANM')
            for fn in fnfilter(names, sys.argv[2]):
                print fn
                extract(dat, fn, sys.argv[3])
        else:
            extract(dat, sys.argv[2], sys.argv[3])