from construct import *
from array import array
import sys

Entry = Struct('entry',
    String('name', 19, padchar=chr(0), paddir='right'),
    OneOf(ULInt8('method'), (0,1)),
    ULInt32('csize'),
    ULInt32('usize'),
    ULInt32('offset')
)

Directory = Struct('directory',
    Magic('SGS.DAT 1.00'),
    ULInt32('file_count'),
    MetaRepeater(lambda ctx: ctx['file_count'], Entry)
)

try:
    # raise ImportError
    from _algos import decompress
except ImportError:
    import warnings
    warnings.warn('could not load _algos, falling back to Python implementation')
    def decompress(string, usize):
        if not string:
            return ''
        data = array('B')
        data.fromstring(string)
        output = array('B')
        data_pos, flags, flag_mask = 0, 0, 0

        while data_pos < len(data):
            if not flag_mask:
                if data_pos + 1 == len(data):
                    raise ValueError
                flags = data[data_pos]
                data_pos += 1
                flag_mask = 0x80

            if flags & flag_mask:
                if data_pos + 1 == len(data):
                    raise ValueError
                ref_in = data[data_pos]
                data_pos += 1
                ref_len = data[data_pos]
                data_pos += 1

                ref_in |= (ref_len & 0x0F) << 8
                ref_in = len(output) - ref_in
                ref_len = ((ref_len & 0xF0) >> 4) + 1
                if ref_in < 0:
                    raise IOError('decompression failure')

                for i in xrange(ref_len):
                    output.append(output[ref_in+i])
            else:
                output.append(data[data_pos])
                data_pos += 1

            flag_mask >>= 1

        return output.tostring()

class DATArchive(object):
    def __init__(self, f, mode='r'):
        if mode not in ('w', 'r'):
            raise ValueError
        self._mode = mode

        if mode == 'r':
            if isinstance(f, basestring):
                self.fn = f
                self.f = f = open(f, 'rb')
            else:
                self.f = f
            self.entries = Directory.parse_stream(f).entry
            self.dict = dict((e.name, e) for e in self.entries)
        else:
            if isinstance(f, basestring):
                self.fn = f
                self.f = None
            else:
                raise ValueError #self.f = f
            self.entries = ListContainer()
            self.dict = dict()
            self.sources = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        if self._mode == 'w':
            with open(self.fn, 'wb') as f:
                entries = self.entries
                count = len(entries)
                directory = Container(file_count=count, entry=entries)
                base_offset = offset = 16 + count * 32
                for entry in entries:
                    entry.offset = offset
                    offset += entry.csize
                Directory.build_stream(directory, f)
                assert f.tell() == base_offset
                for entry, source in zip(entries, self.sources):
                    entry.offset = offset
                    type, source = source
                    if type == 'data':
                        f.write(source)
                    elif type == 'copy':
                        arc, e = source
                        if isinstance(arc, basestring):
                            with DATArchive(arc) as dat:
                                df = dat.f
                                df.seek(e.offset)
                                f.write(df.read(e.csize))
                        else:
                            df = arc.f
                            df.seek(e.offset)
                            f.write(df.read(e.csize))
        else:
            self.f.close()

    def filenames(self):
        for entry in self.entries:
            yield entry.name

    def read(self, filename):
        if self._mode != 'r':
            raise RuntimeError
        entry = self.dict[filename]
        f = self.f
        f.seek(entry.offset)
        if entry.method:
            # print 'decompressing', entry.name
            data = decompress(f.read(entry.csize), entry.usize)
            # print entry.name, 'decompressed'
            assert entry.usize == len(data)
            return data
        else:
            return f.read(entry.csize)

    def writestr(self, arcname, bytes):
        if self._mode != 'w':
            raise RuntimeError
        self.entries.append(Container(name=arcname, method=0, csize=len(bytes), usize=len(bytes), offset=0))
        self.sources.append(('data', bytes))

    def copy(self, archive, entry):
        if self._mode != 'w':
            raise RuntimeError
        if isinstance(entry, str):
            if isinstance(archive, basestring):
                with DATArchive(archive) as dat:
                    entry = dat.dict[entry]
            else:
                entry = archive.dict[entry]
        self.entries.append(Container(name=entry.name, method=entry.method, csize=entry.csize, usize=entry.usize, offset=0))
        self.sources.append(('copy', (archive, entry)))

if __name__ == '__main__':
    import os, sys
    from traceback import print_exc

    with DATArchive(sys.argv[1]) as dat:
        match = None
        if len(sys.argv) > 3:
            from fnmatch import fnmatch
            match = sys.argv[3]
        dest = sys.argv[2]
        for fn in dat.filenames():
            if match and not fnmatch(fn, match):
                continue
            path = os.path.join(dest, fn)
            dirpath = os.path.dirname(path)
            if not os.path.exists(dirpath):
                os.makedirs(dirpath)
            print fn
            try:
                data = dat.read(fn)
            except IOError:
                print_exc()
            else:
                with open(path, 'wb') as f:
                    f.write(data)