# encoding=utf_8
import struct
from collections import defaultdict, namedtuple
from random import randint
from construct import Container
from array import array
from warnings import warn
import numpy as np

class IncompleteWarning(UserWarning):
    pass
warnedincomplete = set()
def incomplete(message=''):
    if message in warnedincomplete:
        return
    warn(message, IncompleteWarning, 2)
    warnedincomplete.add(message)

FLAG_LOOP           = 0x0001
FLAG_TRANSPARENCY   = 0x0002
FLAG_SHIMMER        = 0x0004
FLAG_REPEATLAST     = 0x0008 # ?
FLAG_MOTION         = 0x0010 # ?
FLAG_MOTION_LOOP    = 0x0020 # ?
FLAG_MOTION_BOUNCE  = 0x0040 # ?
FLAG_MOTION_POINT   = 0x0080 # ?
FLAG_WAVE           = 0x0100 # ?
FLAG_RANDOM         = 0x4000 # ?
FLAG_INITIALIZED    = 0x8000

TERM_FLAG_MASK      = ~(FLAG_LOOP|FLAG_RANDOM)

STATE_1     = 1 # wait for buttondown or non-escape or -skip keypress
STATE_2     = 2 # wait for a period of time
STATE_3     = 3 # wait for a particular layer to finish animation
STATE_4     = 4 # wait for a fade to completed/started? - also cleared in map movement...
STATE_5     = 5 # waits for things like fades and layer clears clears to be applied
STATE_6     = 6 # waits for a particular layer to finish motion
STATE_7     = 7 # waits for a sound to be deleted?
STATE_8     = 8 # wait for text to be displayed for a period of time based on its length
STATE_9     = 9 # set when painting choices - not explicitly unset anywhere though. clearChoices does unset it among others
STATE_10    = 10 # wait for directional input - possibly only keys
STATE_11    = 11 # wait for there to be no keyboard or mouse input
STATE_12    = 12 # wait for the map movement to be finished
STATE_13    = 13 # wait for a button/key to be pressed (or a particular layer to be deleted)
STATE_14    = 14 # wait for a name to be entered
STATE_15    = 15 # wait for the character to be oriented?

INPUT_UP        = 0x01
INPUT_DOWN      = 0x02
INPUT_LEFT      = 0x04
INPUT_RIGHT     = 0x08
INPUT_ESCAPE    = 0x10
INPUT_SKIP      = 0x20

CHAR_TEXTSUBST = u'\1'
CHAR_TEXTCLEAR = u'＾' # 0x814F \uFF3E clear text
CHAR_TEXTCRLF  = u'￥' # 0x818F \uFFE5 line break
CHAR_TEXTHERO  = u'＄' # 0x8190 \uFF04 ＄ Replace with player name (default: 中田)

DEFAULT_HERO = u'中田'

SYSTEM_COLORS_START = (
    (0x00, 0x00, 0x00),
    (0x80, 0x00, 0x00),
    (0x00, 0x80, 0x00),
    (0x80, 0x80, 0x00),
    (0x00, 0x00, 0x80),
    (0x80, 0x00, 0x80),
    (0x00, 0x80, 0x80),
    (0xC0, 0xC0, 0xC0),
    (0xC0, 0xDC, 0xC0),
    (0xA4, 0xC8, 0xF0))
SYSTEM_COLORS_END = (
    (0xFF, 0xFB, 0xF0),
    (0xA0, 0xA0, 0xA4),
    (0x80, 0x80, 0x80),
    (0xFF, 0x00, 0x00),
    (0x00, 0xFF, 0x00),
    (0xFF, 0xFF, 0x00),
    (0x00, 0x00, 0xFF),
    (0xFF, 0x00, 0xFF),
    (0x00, 0xFF, 0xFF),
    (0xFF, 0xFF, 0xFF))

class Interpreter(object):
    def __init__(self, ui, store):
        self.ui = ui
        self.store = store
        self.functions = dict((int(meth.split('_')[2], 16), getattr(self, meth)) for meth in dir(self) if meth.startswith('_op_'))
        self.debug = None
        self.drawing_x2 = 0
        self.drawing_y2 = 0
        self.ops = []
        self.index = 0
        self.loaded_frames = [None] * 8
        # unused update rectangles
        self.anim_layers = [AnimationLayer() for i in xrange(32)]
        self.track_names = [''] * 40
        self.text_lines = [[]] # 25 originally
        self.names = [u''] * 40
        self.globals = np.zeros(256, dtype=np.int16)
        self.lines_i18n = [''] * 25
        # registry entries...
        self.state = 0
        self.callstack = []
        self.bitmap_palette = np.zeros([256, 3], np.uint8)
        self.palette_from = 10
        self.palette_to = 246
        self.master_palette = np.zeros((256, 3), np.uint8)
        self.current_palette = np.zeros((256, 3), np.uint8)
        self.current_palette[:10] = SYSTEM_COLORS_START
        self.current_palette[246:] = SYSTEM_COLORS_END
        self.current_rgb = np.zeros((256, 3), np.int32)
        self.delta_rgb = np.zeros((256, 3), np.int32)
        self.waitforlayer = 0
        self.pause_anim = False
        self.future_pause = False
        self.delay_countdown = 0
        self.typing_var = 0
        self.reading_countdown = 0
        self.reading_factor = 80 # reasonable default
        self.typing_layer = 0
        self.drawing_x = 0
        self.drawing_y = 0
        self.drawing_w = 640
        self.drawing_h = 400
        self.fade_in = 0
        self.fade_color = 0
        self.future_palette_changed = 0
        self.skip_type =0
        self.inputflags = 0
        self.buttonflags = 0
        self.input_var = 0
        self.inputshift = 0
        self.bytes_printed = 0
        self.continuehandler = None
        self.escapehandler = None
        self.exithandler = None
        self.loadhandler = None
        self.repeathandler = None
        self.mouse_caret = 0
        self.mouse_moved = False
        self.synth_mouse_moved = False
        self.mouse_pos_old = None
        self.mouse_pos_new = (0, 0)
        self.mouse_scrolling = False
        self.pause_anim = False
        self.language = 0
        self.dialogue_line = u''
        self.text_char_x = 0
        self.text_char_y = 0
        self.text_char_w = 640
        self.text_char_h = 80
        self.text_x = 0
        self.text_y = 400
        self.text_color = (255, 255, 255)
        self.font_height = 16 # reasonable default
        self.text_nlpadding = 0
        self.hero_name = DEFAULT_HERO
        self.palette_changed = False
        self.caret_pos = (640 - 20, 480 - 20) # resonable default
        self.caret_type = 0
        self.caret_countdown = 0
        self.caret_index = 0
        self.choices_count = 0
        self.choices_var = 0
        self.choices_layout = 0
        self.choices_x = 0
        self.choices_y = 0
        self.choices_w = 0
        self.choices_h = 0
        self.choices_cellw = 0
        self.choices_text = []
        self.current_sound = None
        self.disabled_pcms = 0
        self.music_volume = 0x8000
        self.voice_volume = 0x8000
        self.music_stepto = 0xFFFF
        self.music_volume2 = 0xFFFF
        self.music_delta = 0
        self.last_trackn = 99
        self.noskipsound = 0
        self.use_dlgpop = 0
        self.dlgpop_bmp = None
        self.dlgpop_x = 0
        self.dlgpop_y = 384
        self.dlgpop_w = 640
        self.dlgpop_h = 80
        self.mapmove_orient = 0
        self.walk_sprite_index = 0
        self.walk_index = 0
        self.mapmove_x = 0
        self.mapmove_y = 0
        self.map_width = 0
        self.map_height = 0
        self.trail_index = 0
        self.move_trail = np.zeros((64, 3), dtype=np.int32)
        self.shimmer_type = 0
        self.shimmer_value = 0
        self.shimmer_counter = 8
    def nextstate(self):
        debug = self.debug
        if debug is not None: debug.nostate(self)
        funcs = self.functions
        while not self.state:
            op = self.nextop()
            if debug is not None: debug.nextop(self, op)
            try:
                # print '%4X %02x %d'%(op.address, op.instruction, self.skip_type)
                func = funcs[op.instruction]
            except KeyError:
                self.missingop(op)
            else:
                func(op)
        if debug is not None: debug.nextstate(self, self.state)
        # print self.state
    def nextop(self):
        i = self.index
        op = self.ops[i]
        self.index = i + 1
        # print '%04x'%op.address, op.fourcc
        return op
    def setops(self, ops):
        self.ops = ops
        self.addresses = dict((op.address, i) for i, op in enumerate(ops))
        self.index = 0
    def setaddress(self, address):
        self.index = self.addresses[address]
    def missingop(self, op):
        raise NotImplementedError('unimplemented op', op)

    def fadeout(self, n, color):
        self.fade_color = color
        self.fade_in = False
        self.palchange_type = 1
        self.palchange_countdown = n
        self.palette_changed = True

        master, curpal = self.master_palette, self.current_palette
        currgb, delta = self.current_rgb, self.delta_rgb
        a, b = self.palette_from, self.palette_to

        curpal[a:b] = master[a:b]
        currgb[a:b] = master[a:b]
        currgb[a:b] <<= 8
        delta[a:b] = color
        delta[a:b] -= master[a:b]
        delta[a:b] <<= 8
        delta[a:b] /= n
    def fadein(self, n, color):
        self.fade_color = color
        self.fade_in = True
        self.palchange_type = 1
        self.palchange_countdown = n
        self.palette_changed = True

        master, curpal = self.master_palette, self.current_palette
        currgb, delta = self.current_rgb, self.delta_rgb
        a, b = self.palette_from, self.palette_to

        curpal[a:b] = color
        currgb[a:b] = curpal[a:b]
        currgb[a:b] <<= 8
        delta[a:b] = master[a:b]
        delta[a:b] -= curpal[a:b]
        delta[a:b] <<= 8
        delta[a:b] /= n
    def changepalette(self):
        self.bitmap_palette[10:246] = self.current_palette[10:246]
        ptype, index = self.palchange_type, self.palchange_countdown
        if ptype == 0:
            self.palette_changed = False
            if self.state == STATE_4:
                self.ui.clearstate()
        elif ptype == 1:
            if index:
                self.stepfade()
                self.palchange_countdown = index - 1
            else:
                self.palchange_type = 0
        elif ptype == 2:
            if index:
                self.stepfade()
                index -= 1
            else:
                a, b = self.palette_from, self.palette_to
                self.delta_rgb[a:b] *= -1
                index = self.palchange_count
            self.palchange_countdown = index
        elif ptype == 3:
            index -= 1
            if not index:
                self.stepfade()
                a, b = self.palette_from, self.palette_to
                self.delta_rgb[a:b] *= -1
                index = self.palchange_count
            self.palchange_countdown = index
        elif ptype == 4:
            index -= 1
            if not index:
                rgbindex, rgbshift = self.rgbshift_index, self.rgbshift
                index, colors = rgbshift[rgbindex]
                self.current_palette[colors[:, 0]] = colors[:, 1:]
                rgbindex += 1
                if rgbindex >= len(rgbshift):
                    rgbindex = 0
                self.rgbshift_index = rgbindex
            self.palchange_countdown = index
    def stepfade(self):
        curpal = self.current_palette
        currgb, delta = self.current_rgb, self.delta_rgb
        a, b = self.palette_from, self.palette_to
        currgb[a:b] += delta[a:b]
        curpal[a:b] = np.clip(currgb[a:b] >> 8, 0, 255)
    def endfade(self):
        a, b = self.palette_from, self.palette_to
        if self.fade_in:
            self.current_palette[a:b] = self.master_palette[a:b]
        else:
            self.current_palette[a:b] = self.fade_color

    def _op_00_terminate(self, op):
        raise RuntimeError('encountered terminator')
    def _op_01(self, op):
        self.drawing_x2 = op.x
        self.drawing_y2 = op.y
    def _op_02(self, op):
        self.drawing_x = self.drawing_x2
        self.drawing_y = self.drawing_y2
        self.drawing_w = op.w
        self.drawing_h = op.h
    def _op_03_loadanm(self, op):
        self.loaded_frames[op.a] = self.store.loadanm(op.path)
    def _op_04_freeanm(self, op):
        self.loaded_frames[op.a] = None
    def _op_05_fadein(self, op):
        self.fadein(op.div, (op.a, op.a, op.a))
        self.state = STATE_4
    def _op_06_fadeout(self, op):
        self.fadeout(op.div, (op.a, op.a, op.a))
        self.state = STATE_4
    def _op_07_loadpalette(self, op):
        frames = self.loaded_frames[op.a]
        colors = np.asarray([(f.red, f.green, f.blue) for f in frames.palette], np.uint8)
        a, b = self.palette_from, self.palette_to
        self.bitmap_palette[a:b] = colors[a:b]
        self.master_palette[a:b] = colors[a:b]
        self.endfade()
    def _op_08(self, op):
        self.fade_in = True
        self.palette_type = 0
        self.palette_changed = False
        self.state = STATE_5
    def _op_09(self, op):
        self.fade_in = False
        self.palette_type = 0
        self.palette_changed = False
        self.state = STATE_5
    def _op_0A_loadseq(self, op):
        anim_layer = self.anim_layers[op.a]
        anim_layer.sequence_x = self.drawing_x2 + op.b
        anim_layer.sequence_y = self.drawing_y2 + op.c
        anim_layer.frame_count = 1
        anim_layer.sequence_index = 0
        anim_layer.flags = op.d & 0xFFFF
        anim_layer.anm_index = op.d >> 16 & 0xFF
        anim_layer.mask_color = op.d >> 24 & 0xFF
        anim_layer.sequences = [SequenceItem(e.frame, e.b) for e in op.e]
        self.future_palette_changed = self.fade_in
    def _op_0B_rmlayer(self, op):
        self.anim_layers[op.level].flags = 0
        self.state = STATE_5
    def _op_4F(self, op):
        anim_layer = self.anim_layers[op.level]
        if anim_layer.flags & 0x8000:
            anim_layer.flags &= TERM_FLAG_MASK
            self.future_pause = True
            self.waitforlayer = op.level
            self.state = STATE_3
    def _op_0C(self, op):
        anim_layer = self.anim_layers[op.level]
        if anim_layer.flags & 0x8000:
            anim_layer.flags &= TERM_FLAG_MASK
            self.waitforlayer = op.level
            self.state = STATE_3
    def _op_0D(self, op):
        for anim_layer in self.anim_layers:
            anim_layer.flags = 0
        self.state = STATE_5
    def _op_0E(self, op):
        if not self.skip_type:
            self.caret_type = 2 if self.mouse_caret else 1
            self.caret_index = 0
            self.caret_countdown = 0
            self.state = STATE_1
    _op_0F = _op_0E
    def _op_10(self, op):
        self.loaded_frames[op.a] = Container(uncompressed=True, frame_count=0, offset_start=772, offsets=[])
    def _op_12_copyimage(self, op):
        frames = self.loaded_frames[op.c]
        frameorig = self.loaded_frames[op.a].offsets[op.b-1].frame.value
        framecopy = Container(x=op.d, y=op.e, w=frameorig.w, h=frameorig.h, data=frameorig.data)
        frames.offsets.append(Container(frame=Container(value=framecopy)))
        frames.frame_count += 1
    def _op_61(self, op):
        frames = self.loaded_frames[op.a]
        w, h = op.w, op.h
        framecopy = Container(x=op.x, y=op.y, w=w, h=h, data='\0'*(w*h))
        frames.offsets.append(Container(frame=Container(value=framecopy)))
        frames.frame_count += 1
    def _op_13_delay(self, op):
        self.delay_countdown = n = 24 * op.a / 100
        if n and not self.skip_type:
            self.state = STATE_2
    def _op_14_jump(self, op):
        self.setaddress(op.target)
    def _op_15_setcontinue(self, op):
        self.continuehandler = op.target
    def _op_16_setescape(self, op):
        self.escapehandler = op.target
    def _op_17_tweenmove(self, op):
        anim_layer = self.anim_layers[op.a]
        anim_layer.vector_curx = op.b << 16
        anim_layer.vector_cury = op.e << 16
        anim_layer.vector_width = op.d
        anim_layer.vector_height = op.g
        anim_layer.vector_deltax = (op.c << 16) - (op.b << 16)
        anim_layer.vector_deltay = (op.f << 16) - (op.e << 16)
        anim_layer.vector_countdown = op.h
        anim_layer.vector_initx = op.b << 16
        anim_layer.vector_inity = op.e << 16
        anim_layer.vector_initcount = op.h
        anim_layer.vector_endx = op.c << 16
        anim_layer.vector_endy = op.f << 16
        if op.h:
            anim_layer.vector_deltax /= op.h
            anim_layer.vector_deltay /= op.h
        else:
            anim_layer.vector_deltax = 0
            anim_layer.vector_deltay = 0
    def _op_18(self, op):
        flags = FLAG_INITIALIZED | FLAG_MOTION
        if self.anim_layers[op.a].flags & flags == flags:
            self.waitforlayer = op.a
            self.state = STATE_6
    def _op_5C(self, op):
        anim_layer = self.anim_layers[op.a]
        anim_layer.vector_countdown = 0
        anim_layer.vector_initcount = 0
    def _op_19(self, op):
        self.music_volume2 = self.music_stepto
        self.music_delta = 0
        self.ui.setvolumes()
        self.ui.playbgm(op.sndnum, op.a)
    def _op_1A(self, op):
        self.music_delta = 0
        self.ui.stopbgm()
    def _op_1B(self, op): 
        self.track_names[op.sndnum] = op.path
    def _op_65(self, op):
        self.last_trackn = op.a
    def _op_5D(self, op):
        self.music_stepto = n = op.volume | (op.volume << 8)
        self.music_volume2 = n
        self.music_delta = 0
        self.ui.setvolumes()
        # print '5D', self.music_delta, self.music_stepto, self.music_volume2
    def _op_5E(self, op):
        self.music_delta = -0xFFFF / op.frames
        self.music_stepto = 0
        # print '5E', self.music_delta, self.music_stepto
    def _op_60(self, op):
        n = op.volume | (op.volume << 8)
        self.music_delta = (n - self.music_stepto) / op.frames
        self.music_stepto = n
        # print '60', self.music_delta, self.music_stepto
    def _op_1D(self, op):
        flags = op.flags
        if not flags & self.disabled_pcms:
            self.ui.loadpcm(op.path, flags & 1)
    def _op_1E(self, op):
        if self.current_sound is not None:
            self.state = STATE_7
            self.ui.deletepcm(1)
    def _op_1F(self, op):
        if self.current_sound is not None:
            self.state = STATE_7
            if self.skip_type == 1:
                self.ui.deletepcm(1)
    def _op_20_textarea(self, op):
        self.text_x = op.x
        self.text_y = op.y
        self.text_char_x = 0
        self.text_char_y = 0
        self.text_char_w = op.w
        self.text_char_h = op.h
    def _op_21(self, op):
        self.bytes_printed = 0
        self.dialogue_line = op.text
        self.ui.drawdialogue()
    def _op_68(self, op):
        self.bytes_printed = 0
        self.lines.append(op.line)
        incomplete('i18n') # dialogue
    def _op_22(self, op):
        self.text_nlpadding = op.a
    def _op_23(self, op):
        self.font_height = op.a
        self.ui.setfontheight(op.a)
    def _op_62(self, op):
        self.use_dlgpop = n = op.a
        if n == 2:
            self.ui.showdialoguepopup(True, update=True)
            self.dlgpop_shown = True
        else:
            self.ui.showdialoguepopup(False)
            if n == 3:
                self.dlgpop_shown = False
    def _op_24_name(self, op):
        self.names[op.name] = op.text
    def _op_25(self, op):
        self.callstack.append(self.ops[self.index].address)
        self.setaddress(op.target)
    def _op_26(self, op):
        self.setaddress(self.callstack.pop())
    def _op_27(self, op):
        self.ui.setfullscreen(op.a)
    def _op_28(self, op):
        self.palchange_type = op.type
        self.palchange_count = op.frames
        self.palchange_countdown = op.frames
        self.palette_changed = True
        divisor = op.frames
        if op.type == 3:
            divisor = 1

        master, curpal = self.master_palette, self.current_palette
        currgb, delta = self.current_rgb, self.delta_rgb

        curpal[10:246] = master[10:246]
        currgb[10:246] = master[10:246]
        currgb[10:246] <<= 8
        delta[10:246] = 0

        colors = np.asarray([(b.i, b.r, b.g, b.b) for b in op.colors], np.uint8)
        index = colors[:, 0]
        colors1 = colors[:, 1:]
        colors2 = master[index]
        curpal[index] = colors2
        currgb[index] = colors2
        currgb[index] <<= 8
        delta[index] = colors1
        delta[index] -= colors2
        delta[index] <<= 8
        delta[index] /= divisor
    def _op_29(self, op):
        self.exithandler = op.target
        self.ui.enableaction('exit')
    def _op_2A(self, op):
        self.exithandler = None
        self.ui.enableaction('exit', False)
    def _op_2B(self, op):
        self.loadhandler = op.target
        self.ui.enableaction('load')
    def _op_2C(self, op):
        self.loadhandler = None
        self.ui.enableaction('load', False)
    def _op_53(self, op):
        self.repeathandler = op.target
        self.ui.enableaction('repeat')
    def _op_54(self, op):
        self.repeathandler = None
        self.ui.enableaction('repeat', False)
    def _op_2D(self, op):
        self.continuehandler = None
        self.escapehandler = None
    def _op_2E(self, op):
        if self.globals[op.a] == op.b:
            self.setaddress(op.target)
    def _op_2F(self, op):
        if self.globals[op.a] != op.b:
            self.setaddress(op.target)
    def _op_30(self, op):
        if self.globals[op.a] <= op.b:
            self.setaddress(op.target)
    def _op_31(self, op):
        if self.globals[op.a] >= op.b:
            self.setaddress(op.target)
    def _op_32(self, op):
        self.switch_value = self.globals[op.a]
    def _op_33(self, op):
        if self.switch_value != op.a:
            self.setaddress(op.target)
    def _op_34(self, op):
        self.globals[op.a] = op.b
    def _op_35(self, op):
        self.globals[op.a] += 1
    def _op_36(self, op):
        self.globals[op.a] -= 1
    def _op_37(self, op):
        self.globals[op.a] += op.b
    def _op_38(self, op):
        self.globals[op.a] |= op.b
    def _op_39(self, op):
        self.globals[op.a] &= op.b
    def _op_3A(self, op):
        self.palchange_countdown = op.a
        self.palette_changed = True
        self.palchange_type = 1
        divisor = op.a

        curpal = self.current_palette
        currgb, delta = self.current_rgb, self.delta_rgb

        currgb[10:246] = curpal[10:246]
        currgb[10:246] <<= 8
        delta[10:246] = currgb[10:246]
        delta[10:246] *= -1
        delta[10:246] <<= 8
        delta[10:246] /= divisor
        
        self.state = STATE_4
    def _op_3B(self, op):
        self.loadglobals("SaveSystem", 0, 16)
    def _op_3C(self, op):
        self.saveglobals("SaveSystem", 0, 16)
    def loadglobals(self, name, start, end):
        data = self.store.loadglobals(name)
        if data is not None and len(data) == end - start:
            self.globals[start:end] = data
    def saveglobals(self, name, start, end):
        self.store.saveglobals(name, self.globals[start:end])
    def _op_3D(self, op):
        self.ui.cleararea(op.x, op.y, op.w, op.h)
        self.state = STATE_5
    def _op_3E(self, op):
        self.skip_type = 0
        self.choices_var = op.a
        self.choices_x = op.b
        self.choices_y = op.c
        self.choices_w = op.d * self.font_height / 2
        self.choices_count = op.n
        self.choices_h = op.n * self.font_height
        ss = [f.text for f in op.f]
        self.choices_text = [self.names[ord(s[1])] if s[0] == '\1' else s for s in ss]
        if self.language >= 1:
            v178 = self.ui.gettextwidth(self.choices_text)
        else:
            v178 = max((len(s.encode('sjis')) for s in self.choices_text)) * self.font_height / 2
        if self.choices_layout < 1: # top-down
            self.choices_w = v178
            self.choices_cellw = v178
        elif self.choices_layout == 1: # left-to-right
            self.choices_cellw = v178 + 16
            self.choices_w = self.choices_count * self.choices_cellw
            self.choices_h = self.font_height
        elif self.choices_layout == 3:
            self.choices_cellw = v178 + 16
            self.choices_w = 3 * self.choices_cellw
            self.choices_h = self.font_height * (self.choices_count + 2) / 3
        elif self.choices_layout == 4:
            self.choices_cellw = v178 + 16
            self.choices_w = 4 * self.choices_cellw
            self.choices_h = self.font_height * (self.choices_count + 3) / 4
        self.ui.paintchoices(True)
        self.state = STATE_9
    def _op_3F(self, op):
        self.choices_layout = op.a
    def _op_40(self, op):
        self.enablesave = b = op.a
        self.ui.enableaction('save', b)
    def _op_41(self, op):
        # assuming i got the order right...
        colors = np.asarray([(b.i, b.r, b.g, b.b) for b in op.b], np.uint8)
        self.master_palette[colors[:, 0], :3] = colors[:, 1:]
        self.bitmap_palette[colors[:, 0]] = colors[:, 1:]
        self.endfade()
    def _op_42(self, op):
        if not self.skip_type and self.reading_countdown:
            self.state = STATE_8
    def _op_43(self, op):
        self.reading_factor = op.a
    def _op_44(self, op):
        self.input_var = op.a
        self.state = STATE_10
    def _op_45(self, op):
        self.skip_type = 0
        self.globals[op.a] = self.inputflags
    def _op_46(self, op):
        if self.inputflags:
            self.state = STATE_11
    def _op_47(self, op):
        self.reset1()
    def reset1(self):
        del self.callstack[:]
        self.bytes_printed = 0
        for anim_layer in anim_layers:
            anim_layer.flags = 0
        self.loaded_frames[:] = [None] * len(self.loaded_frames)
    def resetandnext(self):
        del self.callstack[:]
        self.bytes_printed = 0
        self.continuehandler = None
        self.escapehandler = None
        self.exithandler = None
        self.loadhandler = None
        self.repeathandler = None
        self.mouse_caret = 0
        self.pause_anim = False
        self.ui.clearchoices(True)
        self.ui.clearcaret()
        self.ui.clearstate()
    def _op_48(self, op):
        if not op.a & self.switch_value:
            self.setaddress(op.target)
    def _op_49(self, op):
        self.globals[op.storeat] = self.mouseposcolor(op.a, op.b - 1)
    def mouseposcolor(self, anm_index, image_index):
        color = 255
        if self.mouse_moved or self.synth_mouse_moved:
            image = self.loaded_frames[anm_index].offsets[image_index].frame.value
            x, y, w, h = image.x, image.y, image.w, image.h
            mx, my = self.mouse_pos_new
            if x <= mx and y <= my and mx < x + w and my < y + h:
                color = ord(image.data[(h-my+y-1)*w+mx-x])
            self.mouse_moved = False
        return color
    def _op_4A_loadwin(self, op):
        del self.callstack[:]
        self.bytes_printed = 0
        self.continuehandler = None
        self.escapehandler = None
        self.exithandler = None
        self.loadhandler = None
        self.repeathandler = None
        self.ui.enableaction('exit', False)
        self.ui.enableaction('repeat', False)
        self.ui.enableaction('load', False)
        self.setops(self.store.loadwin(op.path))
        if self.debug is not None: self.debug.setcurrentwin(op.path, self.ops)
    def _op_4B(self, op):
        self.caret_pos = (op.x, op.y)
    def _op_4C(self, op):
        self.scroll_x1 = x1 = op.a
        self.scroll_x2 = x2 = op.b
        self.scroll_y1 = y1 = op.c
        self.scroll_y2 = y2 = op.d
        self.mouse_caret = n = op.b | op.c
        if n:
            flags = FLAG_INITIALIZED | FLAG_MOTION
            for anim_layer in self.anim_layers:
                if anim_layer.flags & flags == flags:
                    if x2:
                        anim_layer.vector_deltax = anim_layer.vector_endx / x2
                    else:
                        anim_layer.vector_deltax = 0
                    anim_layer.vector_curx = x1 * anim_layer.vector_deltax
                    if y2:
                        anim_layer.vector_deltay = anim_layer.vector_endy / y2
                    else:
                        anim_layer.vector_deltay = 0
                    anim_layer.vector_cury = y1 * anim_layer.vector_deltay
    def _op_4D(self, op):
        self.pause_anim = 0
        self.future_pause = 0
    def _op_4E(self, op):
        self.pause_anim = 1
    def _op_50(self, op):
        self.fadein(op.div, (op.r, op.g, op.b))
        self.state = STATE_4
    def _op_51(self, op):
        self.fadeout(op.div, (op.r, op.g, op.b))
        self.state = STATE_4
    def _op_52(self, op):
        self.shimmer_type = n = op.a
        if n == 1:
            self.shimmer_value = 0
            self.shimmer_counter = 8
        elif n == 2:
            self.shimmer_value = 65535
            self.shimmer_counter = 8
    def _op_55(self, op):
        self.palette_from = op.a
        self.palette_to = op.b
    def _op_56(self, op):
        self.loadglobals(op.a, 16, 256)
    def _op_57(self, op):
        self.saveglobals(op.a, 16, 256)
    def _op_58(self, op):
        self.globals[op.a] = self.globals[op.b]
    def _op_59(self, op):
        anim_layer = self.anim_layers[op.a]
        anim_layer.wave_cur = op.b
        anim_layer.wave_deltax = op.d # was * 8, but i 1) separate sine and cosine and 2) address them by int instead of char
        anim_layer.wave_delta = op.c
        anim_layer.wave_deltay = op.e # see above
        anim_layer.wave_countdown = op.f
        if op.f:
            anim_layer.wave_delta = (op.c - op.b) / op.f
    def _op_5A(self, op):
        flags = FLAG_INITIALIZED | FLAG_WAVE
        if self.anim_layers[op.a].flags & flags == flags:
            self.waitforlayer = op.a
            self.state = STATE_12
    def _op_5B(self, op):
        anim_layer = self.anim_layers[op.a]
        anim_layer.point_countdown = op.n
        anim_layer.point_count = op.n
        anim_layer.current_point = 0
        anim_layer.points = [Point(c.a, c.b) for c in op.c]
    def _op_63(self, op):
        self.disabled_pcms = op.flags
    def _op_64(self, op):
        self.palchange_countdown = 1
        self.palchange_type = 4
        self.palette_changed = True
        self.rgbshift_index = 0
        self.rgbshift = [(a.frames, np.asarray([(b.i, b.r, b.g, b.b) for b in a.b], np.uint8)) for a in op.a]
    def _op_66(self, op):
        self.synth_mouse_moved = op.a
    def _op_67(self, op):
        incomplete('i18n') # menu
    def _op_69(self, op):
        self.skip_type = 0
        incomplete('i18n') # choices
    def _op_6A(self, op):
        self.globals[op.a] = randint(0, op.mod - 1)
    def _op_6B(self, op):
        self.ui.composepcm(op.path, 441 * op.delay, op.flags)
    def _op_6C(self, op):
        if not op.a & self.disabled_pcms:
            self.ui.startpcm(op.a & 1)
    def _op_6D(self, op):
        if self.anim_layers[op.a].flags:
            self.setaddress(op.target)
    def _op_6E(self, op):
        self.typing_var = op.a
        self.typing_layer = op.b
        self.state = STATE_13
    def _op_6F(self, op):
        if self.globals[op.a] == self.globals[op.b]:
            self.setaddress(op.target)
    def _op_70(self, op):
        if self.globals[op.a] != self.globals[op.b]:
            self.setaddress(op.target)
    def _op_71(self, op):
        if self.globals[op.a] <= self.globals[op.b]:
            self.setaddress(op.target)
    def _op_72(self, op):
        if self.globals[op.a] >= self.globals[op.b]:
            self.setaddress(op.target)
    def _op_73(self, op): # this is probably used in Typing
        self.dialogue_line = op.a
        self.ui.drawtext()
    def _op_74(self, op):
        self.text_color = (op.r, op.g, op.b)
    def _op_75(self, op):
        self.text_position = (op.x, op.y)
    def _op_76(self, op):
        self.ui.askforname()
        self.state = STATE_14
    def _op_78(self, op):
        self.noskipsound = op.a
    def _op_79_startmap(self, op):
        self.mapvar = op.a
        self.state = STATE_15
    def _op_7A_caseand(self, op):
        if op.a & self.switch_value:
            self.setaddress(op.target)
    def _op_7B(self, op):
        self.dlgpop_bmp = data = self.store.loadbmp(op.path)
        w, h = struct.unpack_from('II', data, 0x12)
        self.dlgpop_x = op.a
        self.dlgpop_y = op.b
        self.dlgpop_w, self.dlgpop_h = w, h
    def _op_5F_nop(self, op):
        return

class AnimationLayer(object):
    __slots__ = (
        'flags',
        'anm_index',
        'mask_color',
        'frame_count',
        'image_index',
        'sequence_index',
        'sequence_x',
        'sequence_y',
        'vector_curx',
        'vector_cury',
        'vector_width',
        'vector_height',
        'vector_deltax',
        'vector_deltay',
        'vector_countdown',
        'vector_initx',
        'vector_inity',
        'vector_initcount',
        'vector_endx',
        'vector_endy',
        'wave_cur',
        'wave_deltax',
        'wave_shift',
        'wave_delta',
        'wave_deltay',
        'wave_countdown',
        'point_countdown',
        'point_count',
        'current_point',
        'sequences', # 27 {short, short}[?]
        'points' # 539 {int, int}[256]
    )
    def __init__(self):
        for name in self.__slots__:
            setattr(self, name, 0)
        self.sequences = []
        self.points = []
    def __str__(self):
        return 'AnimationLayer({0})'.format(', '.join('{0}={1}'.format(name, getattr(self, name)) for name in self.__slots__))

SequenceItem = namedtuple('SequenceItem', 'index count')
Point = namedtuple('Point', 'x y')